from signal import signal, SIGTERM
from types import FrameType
from typing import Optional

from buz.event.strategies.execution_strategy.execution_strategy import ExecutionStrategy


class Worker:
    def __init__(
        self,
        execution_strategy: ExecutionStrategy,
    ) -> None:
        self.__execution_strategy = execution_strategy

    def start(self) -> None:
        signal(SIGTERM, self.__sigterm_handler)
        print("Starting buz worker...")
        self.__execution_strategy.start()
        print("Buz worker stopped gracefully")

    def __sigterm_handler(self, signum: int, frame: Optional[FrameType]) -> None:
        self.stop()

    def stop(self) -> None:
        print("Stopping buz worker...")
        self.__execution_strategy.stop()



from typing import Type, get_type_hints

from buz.command import Command
from buz.command.synchronous.command_handler import CommandHandler


class BaseCommandHandler(CommandHandler):
    @classmethod
    def fqn(cls) -> str:
        return f"command_handler.{cls.__module__}.{cls.__name__}"

    @classmethod
    def handles(cls) -> Type[Command]:
        handle_types = get_type_hints(cls.handle)

        if "command" not in handle_types:
            raise TypeError("command parameter not found in handle method")

        if not issubclass(handle_types["command"], Command):
            raise TypeError("command parameter is not an buz.command.Command subclass")

        return handle_types["command"]



from buz.locator.pypendency.handler_not_found_exception import HandlerNotFoundException
from buz.locator.pypendency.handler_not_registered_exception import HandlerNotRegisteredException
from buz.locator.pypendency.container_locator import ContainerLocator


__all__ = [
    "HandlerNotRegisteredException",
    "HandlerNotFoundException",
    "ContainerLocator",
]



class EventAlreadyInProgressException(Exception):
    def __init__(self, event_id: str, subscriber_fqn: str) -> None:
        self.event_id = event_id
        self.subscriber_fqn = subscriber_fqn
        super().__init__(f"Event {event_id} is already in progress by subscriber {subscriber_fqn}")



from typing import List

from buz.middleware import MiddlewareChainBuilder
from buz.query import Query, QueryResponse
from buz.query.synchronous import QueryHandler
from buz.query.synchronous.middleware import HandleMiddleware, HandleCallable


class HandleMiddlewareChainResolver:
    def __init__(self, middlewares: List[HandleMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            HandleCallable, HandleMiddleware
        ] = MiddlewareChainBuilder(middlewares)

    def resolve(self, query: Query, query_handler: QueryHandler, handle: HandleCallable) -> QueryResponse:
        chain_callable: HandleCallable = self.__middleware_chain_builder.get_chain_callable(
            handle, self.__get_middleware_callable
        )
        return chain_callable(query, query_handler)

    @staticmethod
    def __get_middleware_callable(middleware: HandleMiddleware, handle_callable: HandleCallable) -> HandleCallable:
        return lambda query, query_handler: middleware.on_handle(query, query_handler, handle_callable)



from abc import abstractmethod
from typing import Callable, Awaitable

from buz.command import Command
from buz.command.asynchronous import CommandHandler
from buz.middleware import Middleware

HandleCallable = Callable[[Command, CommandHandler], Awaitable[None]]


class HandleMiddleware(Middleware):
    @abstractmethod
    async def on_handle(self, command: Command, command_handler: CommandHandler, handle: HandleCallable) -> None:
        pass



from abc import abstractmethod
from typing import Callable

from buz.event import Event, Subscriber
from buz.middleware import Middleware

ConsumeCallable = Callable[[Event, Subscriber], None]


class ConsumeMiddleware(Middleware):
    @abstractmethod
    def on_consume(self, event: Event, subscriber: Subscriber, consume: ConsumeCallable) -> None:
        pass



from buz.event.transactional_outbox.event_to_outbox_record_translator import EventToOutboxRecordTranslator
from buz.event.transactional_outbox.fqn_to_event_mapper import FqnToEventMapper
from buz.event.transactional_outbox.outbox_record import OutboxRecord
from buz.event.transactional_outbox.outbox_criteria import (
    OutboxSortingCriteria,
    OutboxCriteria,
    OutboxCriteriaFactory,
    DeliverableRecordsOutboxCriteriaFactory,
)
from buz.event.transactional_outbox.outbox_record_finder import (
    OutboxRecordStreamFinder,
    PollingOutboxRecordStreamFinder,
)
from buz.event.transactional_outbox.outbox_record_to_event_translator import OutboxRecordToEventTranslator
from buz.event.transactional_outbox.outbox_repository import OutboxRepository
from buz.event.transactional_outbox.transactional_outbox_event_bus import TransactionalOutboxEventBus
from buz.event.transactional_outbox.transactional_outbox_worker import TransactionalOutboxWorker

__all__ = [
    "OutboxRecord",
    "OutboxSortingCriteria",
    "OutboxCriteria",
    "OutboxCriteriaFactory",
    "DeliverableRecordsOutboxCriteriaFactory",
    "OutboxRepository",
    "FqnToEventMapper",
    "EventToOutboxRecordTranslator",
    "OutboxRecordToEventTranslator",
    "OutboxRecordStreamFinder",
    "PollingOutboxRecordStreamFinder",
    "TransactionalOutboxEventBus",
    "TransactionalOutboxWorker",
]



from __future__ import annotations

from logging import Logger
from typing import Type, Tuple

from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from kafka.coordinator.assignors.roundrobin import RoundRobinPartitionAssignor

from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.services.kafka_consumer import KafkaConsumer
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_client_factory import KafkaPythonClientFactory
from buz.kafka.infrastructure.kafka_python.kafka_python_consumer import KafkaPythonConsumer
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import JSONByteSerializer
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer


class KafkaPythonConsumerFactory(KafkaPythonClientFactory):
    def __init__(
        self,
        logger: Logger,
        kafka_connection_config: KafkaConnectionConfig,
        kafka_partition_assignors: Tuple[Type[AbstractPartitionAssignor], ...] = (),
        byte_serializer: ByteSerializer = JSONByteSerializer(),
        header_serializer: KafkaHeaderSerializer = KafkaHeaderSerializer(),
    ):
        super().__init__(logger=logger, kafka_connection_config=kafka_connection_config)
        self._kafka_partition_assignors = self.__select_partition_assignors(kafka_partition_assignors)
        self._kafka_connection_config = kafka_connection_config
        self._byte_serializer = byte_serializer
        self._header_serializer = header_serializer

    def __select_partition_assignors(
        self, kafka_partition_assignors: Tuple[Type[AbstractPartitionAssignor], ...]
    ) -> Tuple[Type[AbstractPartitionAssignor], ...]:
        # Order matters. The Consumer will try to use the left-most assignor.
        # The assignor used by the GroupCoordinator is agreed between all consumers in group.
        # The list is used to support rolling-updates.

        default_assignors = [RoundRobinPartitionAssignor]

        partition_assignors: list[Type[AbstractPartitionAssignor]] = list(kafka_partition_assignors)
        partition_assignors.extend(default_assignors)

        return tuple(partition_assignors)

    def build(self) -> KafkaConsumer:
        return KafkaPythonConsumer(
            bootstrap_servers=self._kafka_connection_config.bootstrap_servers,
            client_id=self._kafka_connection_config.client_id,
            partition_assignors=self._kafka_partition_assignors,
            logger=self._logger,
            byte_serializer=self._byte_serializer,
            **self._get_autentication_configuration_by_security_protocol(),  # type: ignore
        )



from buz.event.transactional_outbox.outbox_criteria import OutboxCriteria, OutboxCriteriaFactory, OutboxSortingCriteria


class DeliverableRecordsOutboxCriteriaFactory(OutboxCriteriaFactory):
    def create(self) -> OutboxCriteria:
        return OutboxCriteria(delivered_at=None, order_by=OutboxSortingCriteria.CREATED_AT, delivery_paused=False)



from logging import Logger
from typing import Dict, Set, List, Optional, Callable, cast

from kombu import Connection, Queue, Consumer as MessageConsumer, Message
from kombu.mixins import ConsumerMixin
from kombu.transport.pyamqp import Channel

from buz.event import Event, Subscriber
from buz.event.infrastructure.kombu.allowed_kombu_serializer import AllowedKombuSerializer
from buz.event.exceptions.term_signal_interruption_exception import TermSignalInterruptionException
from buz.event.middleware import ConsumeMiddleware, ConsumeMiddlewareChainResolver
from buz.event.middleware.exceptions.event_already_in_progress_exception import EventAlreadyInProgressException
from buz.locator import Locator
from buz.event.consumer import Consumer
from buz.event.exceptions.event_restore_exception import EventRestoreException
from buz.event.exceptions.subscribers_not_found_exception import SubscribersNotFoundException
from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.reject_callback import RejectCallback

QueueToSubscriberFqnMapping = Dict[Queue, Set[str]]


class KombuConsumer(ConsumerMixin, Consumer):
    def __init__(
        self,
        connection: Connection,
        queues_mapping: QueueToSubscriberFqnMapping,
        serializer: AllowedKombuSerializer,
        prefetch_count: int,
        locator: Locator[Event, Subscriber],
        logger: Logger,
        consume_retrier: Optional[ConsumeRetrier] = None,
        reject_callback: Optional[RejectCallback] = None,
        consume_middlewares: Optional[List[ConsumeMiddleware]] = None,
    ):
        self.connection = connection
        self.__queues_mapping = queues_mapping
        self.__serializer = serializer
        self.__prefetch_count = prefetch_count
        self.__locator = locator
        self.__logger = logger
        self.__consume_retrier = consume_retrier
        self.__reject_callback = reject_callback
        self.__consume_middleware_chain_resolver = ConsumeMiddlewareChainResolver(consume_middlewares or [])

    def stop(self) -> None:
        self.should_stop = True

    def get_consumers(self, consumer_factory: Callable, channel: Channel) -> List[MessageConsumer]:
        return [
            consumer_factory(
                queues=[queue],
                callbacks=self.__get_consumer_callbacks(allowed_subscriber_fqns),
                prefetch_count=self.__prefetch_count,
                accept=[self.__serializer],
            )
            for queue, allowed_subscriber_fqns in self.__queues_mapping.items()
        ]

    def __get_consumer_callbacks(self, allowed_subscriber_fqns: Set[str]) -> List[Callable[[Dict, Message], None]]:
        return [lambda body, message: self.__on_message_received(body, message, allowed_subscriber_fqns)]

    def __on_message_received(self, body: Dict, message: Message, allowed_subscriber_fqns: Set[str]) -> None:
        try:
            event = self.__restore_event(body, message)
            subscribers = self.__subscribers(event, allowed_subscriber_fqns)
        except (EventRestoreException, SubscribersNotFoundException) as exc:
            self.__logger.exception(f"Message cannot be processed: {exc}")
            message.ack()
            return
        except Exception as exc:
            self.__logger.exception(f"Unknown error while processing message: {exc}")
            message.ack()
            return

        self.__consume_event(message, event, subscribers)

    def __restore_event(self, body: Dict, message: Message) -> Event:
        try:
            event_fqn = message.headers["fqn"]
            event_klass = self.__locator.get_message_klass_by_fqn(event_fqn)
            return cast(Event, event_klass.restore(**body))
        except Exception as exc:
            raise EventRestoreException(body, str(message)) from exc

    def __subscribers(self, event: Event, allowed_subscriber_fqns: Set[str]) -> List[Subscriber]:
        event_subscribers = None
        try:
            event_subscribers = self.__locator.get(event)
            allowed_event_subscribers = [
                subscriber for subscriber in event_subscribers if subscriber.fqn() in allowed_subscriber_fqns
            ]
            if len(allowed_event_subscribers) == 0:
                raise SubscribersNotFoundException(event, allowed_subscriber_fqns, event_subscribers)

            return allowed_event_subscribers
        except Exception as exc:
            raise SubscribersNotFoundException(event, allowed_subscriber_fqns, event_subscribers) from exc

    def __consume_event(self, message: Message, event: Event, subscribers: List[Subscriber]) -> None:
        try:
            if self.should_stop is True:
                raise TermSignalInterruptionException()

            for subscriber in subscribers:
                self.__consume_middleware_chain_resolver.resolve(event, subscriber, self.__perform_consume)
            message.ack()
        except Exception as exc:
            self.__on_consume_exception(message, event, subscribers, exc)

    def __perform_consume(self, event: Event, subscriber: Subscriber) -> None:
        subscriber.consume(event)

    def __on_consume_exception(
        self, message: Message, event: Event, subscribers: List[Subscriber], exception: Exception
    ) -> None:
        if isinstance(exception, EventAlreadyInProgressException) or isinstance(
            exception, TermSignalInterruptionException
        ):
            message.requeue()
            return

        self.__logger.warning(
            f"Event {event.id} could not be consumed by subscribers {[subscriber.fqn() for subscriber in subscribers]}:"
            f"{exception}."
        )

        if self.__consume_retrier is None:
            self.__reject_message(message, event, subscribers)
            return

        should_retry = self.__consume_retrier.should_retry(event, subscribers)
        if should_retry is True:
            self.__consume_retrier.register_retry(event, subscribers)
            message.requeue()
            return

        self.__reject_message(message, event, subscribers)

    def __reject_message(self, message: Message, event: Event, subscribers: List[Subscriber]) -> None:
        message.reject()
        if self.__reject_callback is not None:
            self.__reject_callback.on_reject(event, subscribers)



from logging import Logger

from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.services.kafka_admin_test_client import KafkaAdminTestClient
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_admin_client_factory import (
    KafkaPythonAdminClientFactory,
)
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_test_client import KafkaPythonAdminTestClient


class KafkaPythonAdminClientTestFactory(KafkaPythonAdminClientFactory):
    def __init__(
        self,
        logger: Logger,
        kafka_connection_config: KafkaConnectionConfig,
    ):
        super().__init__(logger=logger, kafka_connection_config=kafka_connection_config)

    def build(self) -> KafkaAdminTestClient:
        return KafkaPythonAdminTestClient(
            bootstrap_servers=self._kafka_connection_config.bootstrap_servers,
            client_id=self._kafka_connection_config.client_id,
            logger=self._logger,
            **self._get_autentication_configuration_by_security_protocol(),  # type: ignore
        )



from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_consumer_record import KafkaConsumerRecord
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.domain.models.kafka_topic import KafkaTopic
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.domain.services.kafka_admin_test_client import KafkaAdminTestClient
from buz.kafka.domain.services.kafka_consumer import KafkaConsumer
from buz.kafka.domain.services.kafka_producer import KafkaProducer
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_admin_client_factory import (
    KafkaPythonAdminClientFactory,
)
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_admin_test_client_factory import (
    KafkaPythonAdminClientTestFactory,
)
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_consumer_factory import KafkaPythonConsumerFactory
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_producer_factory import KafkaPythonProducerFactory
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_client import KafkaPythonAdminClient
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_test_client import KafkaPythonAdminTestClient
from buz.kafka.infrastructure.kafka_python.kafka_python_consumer import KafkaPythonConsumer
from buz.kafka.infrastructure.kafka_python.kafka_python_producer import KafkaPythonProducer
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import JSONByteSerializer


__all__ = [
    "KafkaConsumer",
    "KafkaPythonConsumer",
    "KafkaProducer",
    "KafkaPythonProducer",
    "KafkaAdminClient",
    "KafkaAdminTestClient",
    "KafkaPythonAdminClient",
    "KafkaPythonAdminTestClient",
    "KafkaPythonAdminClientFactory",
    "KafkaPythonAdminClientTestFactory",
    "KafkaPythonConsumerFactory",
    "KafkaPythonProducerFactory",
    "KafkaTopicsAlreadyCreatedException",
    "KafkaConsumerRecord",
    "KafkaTopic",
    "KafkaSupportedSecurityProtocols",
    "KafkaConnectionConfig",
    "ByteSerializer",
    "JSONByteSerializer",
    "ConsumerInitialOffsetPosition",
]



from abc import ABC, abstractmethod
from typing import Sequence

from buz.event.transactional_outbox import OutboxRecord
from buz.event.transactional_outbox import OutboxCriteria


class OutboxRepository(ABC):
    @abstractmethod
    def save(self, outbox_record: OutboxRecord) -> None:
        pass

    @abstractmethod
    def find(self, criteria: OutboxCriteria) -> Sequence[OutboxRecord]:
        pass

    @abstractmethod
    def bulk_delete(self, outbox_record_ids: Sequence[str]) -> None:
        pass



import asyncio
from asyncio import AbstractEventLoop


def __get_event_loop() -> AbstractEventLoop:
    try:
        result = asyncio.get_running_loop()
    except RuntimeError:
        result = asyncio.new_event_loop()
    return result


loop = __get_event_loop()



from typing import Callable, Coroutine, Union, TypeVar, Generic

from buz.command import Command
from buz.query import Query
from buz.wrapper.event_loop import loop

HandleCallable = Union[Callable[[Command], Coroutine], Callable[[Query], Coroutine]]
T = TypeVar("T")  # Buz input type, Ex: Command, Query, Event
K = TypeVar("K")  # Buz response type, Ex: QueryResponse, None


class AsyncToSync(Generic[T, K]):
    def __init__(self, func: HandleCallable):
        self.__func = func

    def __call__(self, *args: T) -> K:
        return self.__run_task(self.__func, *args)

    @staticmethod
    def __run_task(func: Callable, *args: T) -> K:
        return loop.run_until_complete(func(*args))



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(frozen=True)
class KafkaConsumerRecord(Generic[T]):
    value: T
    headers: dict[str, str]



from buz.event.infrastructure.kombu.publish_strategy.publish_strategy import PublishStrategy
from buz.event.infrastructure.kombu.publish_strategy.fanout_exchange_per_event_publish_strategy import (
    FanoutExchangePerEventPublishStrategy,
)

__all__ = ["PublishStrategy", "FanoutExchangePerEventPublishStrategy"]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Sequence, Type

from buz import Handler
from buz import Message

K = TypeVar("K", bound=Message)
V = TypeVar("V", bound=Handler)


class Locator(ABC, Generic[K, V]):
    @abstractmethod
    def get(self, message: K) -> Sequence[V]:
        pass

    @abstractmethod
    def get_handler_by_fqn(self, handler_fqn: str) -> V:
        pass

    @abstractmethod
    def get_message_klass_by_fqn(self, message_fqn: str) -> Type[K]:
        pass



from __future__ import annotations


class KafkaTopicsAlreadyCreatedException(Exception):
    def __init__(self, topic_names: list[str]):
        self.topic_names = topic_names

        topics_messages = ",".join(self.topic_names)
        super().__init__(f"The topics ${topics_messages} are already created in the cluster")



from abc import abstractmethod, ABC
from typing import Callable, List

from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_consumer_record import KafkaConsumerRecord

DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING = 500


class KafkaConsumer(ABC):
    @abstractmethod
    def consume(
        self,
        *,
        topics: List[str],
        consumer_group: str,
        consumption_callback: Callable[[KafkaConsumerRecord], None],
        initial_offset_position: ConsumerInitialOffsetPosition = ConsumerInitialOffsetPosition.END,
        number_of_messages_to_polling: int = DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING,
    ) -> None:
        pass

    def request_stop(self) -> None:
        """Request a graceful stop
        This method does not stop the consumer in a instantaneous way,
        it will finalize when the current task will be completed (eventually)
        """
        pass



from abc import abstractmethod

from buz.event import Event, Subscriber
from buz.event.middleware import ConsumeMiddleware, ConsumeCallable


class BaseConsumeMiddleware(ConsumeMiddleware):
    def on_consume(self, event: Event, subscriber: Subscriber, consume: ConsumeCallable) -> None:
        self._before_on_consume(event, subscriber)
        consume(event, subscriber)
        self._after_on_consume(event, subscriber)

    @abstractmethod
    def _before_on_consume(self, event: Event, subscriber: Subscriber) -> None:
        pass

    @abstractmethod
    def _after_on_consume(self, event: Event, subscriber: Subscriber) -> None:
        pass



from abc import abstractmethod
from typing import TypeVar
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.message import Message


T = TypeVar("T", bound=Message)


class MessageToByteSerializer(ByteSerializer[T]):
    @abstractmethod
    def serialize(self, data: T) -> bytes:
        pass

    @abstractmethod
    def unserialize(self, data: bytes) -> T:
        pass



from abc import abstractmethod

from buz.query import Query, QueryResponse
from buz.query.asynchronous.query_handler import QueryHandler
from buz.query.asynchronous.middleware.handle_middleware import HandleCallable, HandleMiddleware


class BaseHandleMiddleware(HandleMiddleware):
    async def on_handle(self, query: Query, query_handler: QueryHandler, handle: HandleCallable) -> QueryResponse:
        self._before_on_handle(query, query_handler)
        query_response = await handle(query, query_handler)
        return self._after_on_handle(query, query_handler, query_response)

    @abstractmethod
    def _before_on_handle(self, query: Query, query_handler: QueryHandler) -> None:
        pass

    @abstractmethod
    def _after_on_handle(
        self, query: Query, query_handler: QueryHandler, query_response: QueryResponse
    ) -> QueryResponse:
        pass



from typing import Optional
from buz.event import Event
from buz.event.infrastructure.buz_kafka.publish_strategy.publish_strategy import KafkaPublishStrategy


class TopicPerEventKafkaPublishStrategy(KafkaPublishStrategy):
    def __init__(self, prefix: Optional[str]):
        self._prefix = f"{prefix}." or ""

    def get_topic(self, event: Event) -> str:
        return f"{self._prefix}{event.fqn()}"



from typing import Callable, Optional

from buz.event import Event

from buz.event.infrastructure.kombu.retry_strategy.publish_retry_policy import PublishRetryPolicy

ErrorCallback = Callable[[Event, Exception, range], None]


# @see https://github.com/celery/kombu/blob/main/kombu/utils/functional.py
class SimplePublishRetryPolicy(PublishRetryPolicy):
    def __init__(
        self,
        max_retries: int = 10,
        interval_start: float = 0,
        interval_step: float = 0.1,
        interval_max: float = 1,
        error_callback: Optional[ErrorCallback] = None,
    ):
        self.__max_retries = max_retries
        self.__interval_start = interval_start
        self.__interval_step = interval_step
        self.__interval_max = interval_max
        self.__error_callback = error_callback

    def max_retries(self, event: Event) -> int:
        return self.__max_retries

    def interval_start(self, event: Event) -> float:
        return self.__interval_start

    def interval_step(self, event: Event) -> float:
        return self.__interval_step

    def interval_max(self, event: Event) -> float:
        return self.__interval_max

    def error_callback(self, event: Event, exc: Exception, interval_range: range) -> None:
        if self.__error_callback is not None:
            self.__error_callback(event, exc, interval_range)



import dataclasses
from uuid import UUID

from buz.event import Event
from buz.event.transactional_outbox.outbox_record import OutboxRecord


class EventToOutboxRecordTranslator:
    def translate(self, event: Event) -> OutboxRecord:
        payload = dataclasses.asdict(event)
        payload.pop("id")
        payload.pop("created_at")
        return OutboxRecord(
            event_id=UUID(event.id),
            event_fqn=event.fqn(),
            created_at=event.parsed_created_at(),
            event_payload=payload,
        )



from dataclasses import asdict
from typing import Collection, List, Optional, Set, Dict

from kombu import Connection, Exchange, Producer
from kombu.entity import PERSISTENT_DELIVERY_MODE

from buz.event import Event, EventBus
from buz.event.infrastructure.kombu.publish_strategy import PublishStrategy
from buz.event.infrastructure.kombu.allowed_kombu_serializer import AllowedKombuSerializer
from buz.event.middleware import (
    PublishMiddleware,
    PublishMiddlewareChainResolver,
)

from buz.event.exceptions.event_not_published_exception import EventNotPublishedException
from buz.event.infrastructure.kombu.retry_strategy.publish_retry_policy import PublishRetryPolicy
from buz.event.infrastructure.kombu.retry_strategy.simple_publish_retry_policy import SimplePublishRetryPolicy


class KombuEventBus(EventBus):
    def __init__(
        self,
        connection: Connection,
        publish_strategy: PublishStrategy,
        publish_retry_policy: PublishRetryPolicy = SimplePublishRetryPolicy(),
        serializer: Optional[AllowedKombuSerializer] = AllowedKombuSerializer.JSON,
        publish_middlewares: Optional[List[PublishMiddleware]] = None,
    ):
        self.__connection = connection
        self.__publish_strategy = publish_strategy
        self.__publish_retry_policy = publish_retry_policy
        self.__serializer = serializer
        self.__publish_middleware_chain_resolver = PublishMiddlewareChainResolver(publish_middlewares or [])
        self.__declared_exchanges: Set[Exchange] = set()
        self.__producer: Optional[Producer] = None

    def publish(self, event: Event) -> None:
        self.__publish_middleware_chain_resolver.resolve(event, self.__perform_publish)

    def __perform_publish(self, event: Event) -> None:
        try:
            event_fqn = event.fqn()
            exchange = self.__get_exchange(event_fqn)
            routing_key = self.__publish_strategy.get_routing_key(event_fqn)

            producer = self.__get_producer()

            body = self.__get_body(event)
            headers = self.__get_headers(event)

            producer.publish(
                body,
                exchange=exchange,
                routing_key=routing_key,
                retry=True,
                retry_policy=self.__get_publish_retry_policy_for_event(event),
                headers=headers,
                delivery_mode=PERSISTENT_DELIVERY_MODE,
            )
        except Exception as exc:
            raise EventNotPublishedException(event) from exc

    def __get_publish_retry_policy_for_event(self, event: Event) -> Dict:
        return {
            "max_retries": self.__publish_retry_policy.max_retries(event),
            "interval_start": self.__publish_retry_policy.interval_start(event),
            "interval_step": self.__publish_retry_policy.interval_step(event),
            "interval_max": self.__publish_retry_policy.interval_max(event),
            "errback": lambda exc, interval_range: self.__publish_retry_policy.error_callback(
                event, exc, interval_range
            ),
        }

    def __get_exchange(self, event_fqn: str) -> Exchange:
        exchange = self.__publish_strategy.get_exchange(event_fqn)

        if exchange not in self.__declared_exchanges:
            self.__declare_exchange(exchange)

        return exchange

    def __declare_exchange(self, exchange: Exchange) -> None:
        auto_retry_declare = self.__connection.autoretry(exchange.declare)
        auto_retry_declare()
        self.__declared_exchanges.add(exchange)

    def __get_producer(self) -> Producer:
        if self.__producer is None:
            self.__producer = self.__connection.Producer(serializer=self.__serializer)

        return self.__producer

    def __get_body(self, event: Event) -> Dict:
        return asdict(event)

    def __get_headers(self, event: Event) -> Dict:
        return {"fqn": event.fqn()}

    def bulk_publish(self, events: Collection[Event]) -> None:
        for event in events:
            self.publish(event)



from typing import List

from buz.event import Event, Subscriber

from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.consumed_event_retry_repository import ConsumedEventRetryRepository
from buz.event.strategies.retry.max_retries_negative_exception import InvalidMaxRetriesParamException


class MaxRetriesConsumeRetrier(ConsumeRetrier):
    def __init__(
        self,
        consumed_event_retry_repository: ConsumedEventRetryRepository,
        max_retries: int = 1,
    ):
        self.__consumed_event_retry_repository = consumed_event_retry_repository
        self.__max_retries = max_retries
        self.__check_max_retries_positive()

    def __check_max_retries_positive(self) -> None:
        if self.__max_retries < 0:
            raise InvalidMaxRetriesParamException(self.__max_retries)

    def should_retry(self, event: Event, subscribers: List[Subscriber]) -> bool:
        consumed_event_retry = self.__consumed_event_retry_repository.find_one_by_event_and_subscriber(
            event, subscribers
        )
        return consumed_event_retry.retries < self.__max_retries

    def register_retry(self, event: Event, subscribers: List[Subscriber]) -> None:
        consumed_event_retry = self.__consumed_event_retry_repository.find_one_by_event_and_subscriber(
            event, subscribers
        )
        consumed_event_retry.register_retry()
        self.__consumed_event_retry_repository.save(consumed_event_retry)



from buz.locator.sync.handler_already_registered_exception import HandlerAlreadyRegisteredException
from buz.locator.sync.handler_not_registered_exception import HandlerNotRegisteredException
from buz.locator.sync.instance_locator import InstanceLocator

__all__ = [
    "HandlerAlreadyRegisteredException",
    "HandlerNotRegisteredException",
    "InstanceLocator",
]



from typing import Set, Optional, Sequence

from buz.event import Event, Subscriber


class SubscribersNotFoundException(Exception):
    def __init__(
        self, event: Event, allowed_subscriber_fqns: Set[str], event_subscribers: Optional[Sequence[Subscriber]] = None
    ) -> None:
        self.event = event
        self.allowed_subscriber_fqns = allowed_subscriber_fqns
        self.event_subscribers = event_subscribers

        message = f"Subscribers not found for event {event}. Allowed subscriber FQNs: {allowed_subscriber_fqns}."
        if event_subscribers is not None:
            message += f" All event subscribers found: {event_subscribers}"

        super().__init__(message)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols


@dataclass(frozen=True)
class KafkaConnectionConfig:
    bootstrap_servers: list[str]
    client_id: str
    security_protocol: KafkaSupportedSecurityProtocols
    user: Optional[str]
    password: Optional[str]



from abc import ABC, abstractmethod
from typing import List

from buz.event import Event, Subscriber

from buz.event.strategies.retry.consumed_event_retry import ConsumedEventRetry


class ConsumedEventRetryRepository(ABC):
    @abstractmethod
    def save(self, consumed_event_retry: ConsumedEventRetry) -> None:
        pass

    @abstractmethod
    def find_one_by_event_and_subscriber(self, event: Event, subscribers: List[Subscriber]) -> ConsumedEventRetry:
        pass



from logging import Logger
from typing import List, Optional, Tuple, Type, TypeVar

from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor

from buz.event import Event, Subscriber
from buz.event.consumer import Consumer
from buz.event.infrastructure.buz_kafka.consume_strategy.consume_strategy import KafkaConsumeStrategy
from buz.event.infrastructure.buz_kafka.consume_strategy.kafka_on_fail_strategy import KafkaOnFailStrategy
from buz.event.middleware.consume_middleware import ConsumeMiddleware
from buz.event.middleware.consume_middleware_chain_resolver import ConsumeMiddlewareChainResolver
from buz.event.strategies.retry.consume_retrier import ConsumeRetrier
from buz.event.strategies.retry.reject_callback import RejectCallback
from buz.serializer.message_to_bytes_serializer import MessageToByteSerializer
from buz.serializer.message_to_json_bytes_serializer import MessageClassToJSONByteSerializer
from buz.kafka import (
    KafkaPythonConsumerFactory,
    KafkaConnectionConfig,
    KafkaConsumerRecord,
    ConsumerInitialOffsetPosition,
)


T = TypeVar("T", bound=Event)


class BuzKafkaConsumer(Consumer):
    def __init__(
        self,
        *,
        connection: KafkaConnectionConfig,
        consume_strategy: KafkaConsumeStrategy,
        on_fail_strategy: KafkaOnFailStrategy,
        kafka_partition_assignors: Tuple[Type[AbstractPartitionAssignor], ...] = (),
        subscriber: Subscriber,
        logger: Logger,
        consumer_initial_offset_position: ConsumerInitialOffsetPosition,
        serializer: Optional[MessageToByteSerializer[T]],
        consume_middlewares: Optional[List[ConsumeMiddleware]] = None,
        consume_retrier: Optional[ConsumeRetrier] = None,
        reject_callback: Optional[RejectCallback] = None,
    ):
        # to type this strictly we have to create subscribers with generics
        self.__serializer: MessageToByteSerializer[T] = serializer or MessageClassToJSONByteSerializer(  # type: ignore
            event_class=subscriber.handles()  # type: ignore
        )

        self.__logger = logger
        self.__consume_retrier = consume_retrier
        self.__reject_callback = reject_callback
        self.__subscriber = subscriber
        self.__consumer_initial_offset_position = consumer_initial_offset_position
        self.__consume_middleware_chain_resolver = ConsumeMiddlewareChainResolver(consume_middlewares or [])
        self.__consume_strategy = consume_strategy
        self.__on_fail_strategy = on_fail_strategy

        kafka_python_consumer_factory = KafkaPythonConsumerFactory(
            byte_serializer=self.__serializer,
            kafka_connection_config=connection,
            logger=self.__logger,
            kafka_partition_assignors=kafka_partition_assignors,
        )

        self.__consumer = kafka_python_consumer_factory.build()

    def run(self) -> None:
        self.__consumer.consume(
            consumer_group=self.__consume_strategy.get_subscription_group(self.__subscriber),
            initial_offset_position=self.__consumer_initial_offset_position,
            topics=self.__consume_strategy.get_topics(self.__subscriber),
            consumption_callback=self.consumption_callback,
        )

    def consumption_callback(self, message: KafkaConsumerRecord[T]) -> None:
        self.__consume_middleware_chain_resolver.resolve(message.value, self.__subscriber, self.__perform_consume)

    def stop(self) -> None:
        self.__consumer.request_stop()

    def __perform_consume(self, event: T, subscriber: Subscriber) -> None:
        should_retry = True
        while should_retry is True:
            try:
                return subscriber.consume(event)
            except Exception as exc:
                self.__logger.warning(f"Event {event.id} could not be consumed by the subscriber {subscriber.fqn}")

                should_retry = self.__should_retry(event, subscriber)

                if should_retry is True:
                    self.__register_retry(event, subscriber)
                    continue

                if should_retry is False:
                    if self.__reject_callback:
                        self.__reject_callback.on_reject(event=event, subscribers=[subscriber])

                    if self.__on_fail_strategy == KafkaOnFailStrategy.STOP_ON_FAIL:
                        raise exc

    def __should_retry(self, event: Event, subscriber: Subscriber) -> bool:
        if self.__consume_retrier is None:
            return False

        return self.__consume_retrier.should_retry(event, [subscriber])

    def __register_retry(self, event: Event, subscriber: Subscriber) -> None:
        if self.__consume_retrier is None:
            return None

        return self.__consume_retrier.register_retry(event, [subscriber])



from buz.event.consumer import Consumer
from buz.event.strategies.execution_strategy.execution_strategy import ExecutionStrategy


class SelfProcessExecutionStrategy(ExecutionStrategy):
    def __init__(self, consumer: Consumer):
        self.__consumer = consumer

    def start(self) -> None:
        self.__consumer.run()

    def stop(self) -> None:
        self.__consumer.stop()



from buz.event.event import Event
from buz.event.subscriber import Subscriber
from buz.event.base_subscriber import BaseSubscriber
from buz.event.event_bus import EventBus

__all__ = [
    "Event",
    "Subscriber",
    "BaseSubscriber",
    "EventBus",
]



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import ClassVar, Union, Tuple

from buz.event.transactional_outbox.outbox_criteria import OutboxSortingCriteria


@dataclass(frozen=True)
class OutboxCriteria:
    UNSET_VALUE: ClassVar[object] = object()

    delivered_at: Union[datetime, None, object] = UNSET_VALUE
    delivered_at_previous_to: Union[datetime, object] = UNSET_VALUE
    delivery_paused: Union[bool, object] = UNSET_VALUE
    order_by: Union[OutboxSortingCriteria, None, object] = UNSET_VALUE
    created_within_timerange: Union[Tuple[Union[timedelta, None], timedelta], object] = UNSET_VALUE



from abc import ABC, abstractmethod

from buz.event.transactional_outbox import OutboxRecord


class OutboxRecordValidator(ABC):
    @abstractmethod
    def set_next_validator(self, validator: "OutboxRecordValidator") -> None:
        pass

    @abstractmethod
    def validate(self, record: OutboxRecord) -> None:
        pass



from typing import Type, get_type_hints

from buz.event import Event
from buz.event import Subscriber


class BaseSubscriber(Subscriber):
    @classmethod
    def fqn(cls) -> str:
        return f"subscriber.{cls.__module__}.{cls.__name__}"

    @classmethod
    def handles(cls) -> Type[Event]:
        consume_types = get_type_hints(cls.consume)

        if "event" not in consume_types:
            raise TypeError("event parameter not found in consume method")

        if not issubclass(consume_types["event"], Event):
            raise TypeError("event parameter is not an buz.event.Event subclass")

        return consume_types["event"]



from typing import Optional

from kombu import Exchange

from buz.event.infrastructure.kombu.publish_strategy import PublishStrategy


class FanoutExchangePerEventPublishStrategy(PublishStrategy):
    def get_exchange(self, event_fqn: str) -> Exchange:
        return Exchange(event_fqn, "fanout", durable=True)

    def get_routing_key(self, event_fqn: str) -> Optional[str]:
        return None



from abc import abstractmethod

from buz.command import Command
from buz.command.asynchronous import CommandHandler
from buz.command.asynchronous.middleware.handle_middleware import HandleMiddleware, HandleCallable


class BaseHandleMiddleware(HandleMiddleware):
    async def on_handle(self, command: Command, command_handler: CommandHandler, handle: HandleCallable) -> None:
        self._before_on_handle(command, command_handler)
        await handle(command, command_handler)
        self._after_on_handle(command, command_handler)

    @abstractmethod
    def _before_on_handle(self, command: Command, command_handler: CommandHandler) -> None:
        pass

    @abstractmethod
    def _after_on_handle(self, command: Command, command_handler: CommandHandler) -> None:
        pass



from buz.query.asynchronous.middleware.handle_middleware_chain_resolver import HandleMiddlewareChainResolver
from buz.query.asynchronous.middleware.base_handle_middleware import BaseHandleMiddleware
from buz.query.asynchronous.middleware.handle_middleware import HandleCallable, HandleMiddleware


__all__ = ["BaseHandleMiddleware", "HandleCallable", "HandleMiddleware", "HandleMiddlewareChainResolver"]



from typing import Type, get_type_hints

from buz.query import Query
from buz.query.synchronous.query_handler import QueryHandler


class BaseQueryHandler(QueryHandler):
    @classmethod
    def fqn(cls) -> str:
        return f"query_handler.{cls.__module__}.{cls.__name__}"

    @classmethod
    def handles(cls) -> Type[Query]:
        handle_types = get_type_hints(cls.handle)

        if "query" not in handle_types:
            raise TypeError("query parameter not found in handle method")

        if not issubclass(handle_types["query"], Query):
            raise TypeError("query parameter is not an buz.query.Query subclass")

        return handle_types["query"]



from typing import Optional

from buz.event.transactional_outbox import OutboxRecord
from buz.event.transactional_outbox.outbox_record_validation.outbox_record_validator import OutboxRecordValidator


class AbstractOutboxRecordValidator(OutboxRecordValidator):
    _next_validator: Optional[OutboxRecordValidator] = None

    def set_next_validator(self, validator: OutboxRecordValidator) -> None:
        self._next_validator = validator

    def validate(self, record: OutboxRecord) -> None:
        if self._next_validator is not None:
            self._next_validator.validate(record)

        return None



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Generic, Optional, TypeVar

T = TypeVar("T")


class KafkaProducer(ABC, Generic[T]):
    @abstractmethod
    def produce(
        self,
        *,
        topic: str,
        message: T,
        partition_key: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        pass



from __future__ import annotations

from abc import abstractmethod, ABC

from buz.kafka.domain.models.kafka_topic import KafkaTopic

DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING = 999


class KafkaAdminClient(ABC):
    @abstractmethod
    def create_topics(
        self,
        *,
        topics: set[KafkaTopic],
    ) -> None:
        pass

    @abstractmethod
    def delete_topics(
        self,
        *,
        topics: set[str],
    ) -> None:
        pass

    @abstractmethod
    def get_topics(
        self,
    ) -> set[str]:
        pass



from abc import ABC, abstractmethod
from typing import Collection

from buz.event import Event


class EventBus(ABC):
    @abstractmethod
    def publish(self, event: Event) -> None:
        pass

    @abstractmethod
    def bulk_publish(self, events: Collection[Event]) -> None:
        pass



from dataclasses import dataclass
from typing import List


@dataclass
class ConsumedEventRetry:
    event_id: str
    subscribers_fqns: List[str]
    retries: int

    def register_retry(self) -> None:
        self.retries += 1



from __future__ import annotations

from logging import Logger
from typing import Generic, List, Optional, TypeVar

from kafka import KafkaProducer

from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer


T = TypeVar("T")


class KafkaPythonProducer(KafkaProducer, Generic[T]):
    def __init__(
        self,
        *,
        bootstrap_servers: List[str],
        client_id: str,
        logger: Logger,
        byte_serializer: ByteSerializer[T],
        security_protocol: KafkaSupportedSecurityProtocols,
        sasl_mechanism: Optional[str] = None,
        sasl_plain_username: Optional[str] = None,
        sasl_plain_password: Optional[str] = None,
        retries: int = 0,
        retry_backoff_ms: int = 100,
    ):
        self._logger = bootstrap_servers
        self.__byte_serializer = byte_serializer
        self.__header_serializer = KafkaHeaderSerializer()

        self.__kafkaProducer = KafkaProducer(
            client_id=client_id,
            bootstrap_servers=bootstrap_servers,
            security_protocol=security_protocol.value,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_username=sasl_plain_username,
            sasl_plain_password=sasl_plain_password,
            retries=retries,
            retry_backoff_ms=retry_backoff_ms,
        )

    def produce(
        self,
        *,
        topic: str,
        message: T,
        partition_key: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ) -> None:
        serialized_headers = self.__header_serializer.serialize(headers) if headers is not None else None

        self.__kafkaProducer.send(
            topic=topic,
            value=self.__byte_serializer.serialize(message),
            headers=serialized_headers,
            key=partition_key,
        )



from typing import List, Optional

from buz.command.synchronous import CommandBus, CommandHandler
from buz.command.synchronous.middleware import HandleMiddleware, HandleMiddlewareChainResolver
from buz.command import MoreThanOneCommandHandlerRelatedException, Command
from buz.locator import Locator


class SelfProcessCommandBus(CommandBus):
    def __init__(
        self,
        locator: Locator[Command, CommandHandler],
        middlewares: Optional[List[HandleMiddleware]] = None,
    ):
        self.__locator = locator
        self.__handle_middleware_chain_resolver = HandleMiddlewareChainResolver(middlewares or [])

    def handle(self, command: Command) -> None:
        command_handlers = self.__locator.get(command)

        if len(command_handlers) > 1:
            raise MoreThanOneCommandHandlerRelatedException(command, command_handlers)

        command_handler = command_handlers[0]
        self.__handle_middleware_chain_resolver.resolve(command, command_handler, self.__perform_handle)

    @staticmethod
    def __perform_handle(command: Command, command_handler: CommandHandler) -> None:
        command_handler.handle(command)



from typing import List

from buz.event import Event, Subscriber
from buz.event.middleware import ConsumeMiddleware, ConsumeCallable
from buz.middleware import MiddlewareChainBuilder


class ConsumeMiddlewareChainResolver:
    def __init__(self, middlewares: List[ConsumeMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            ConsumeCallable, ConsumeMiddleware
        ] = MiddlewareChainBuilder(middlewares)

    def resolve(self, event: Event, subscriber: Subscriber, consume: ConsumeCallable) -> None:
        chain_callable: ConsumeCallable = self.__middleware_chain_builder.get_chain_callable(
            consume, self.__get_middleware_callable
        )
        chain_callable(event, subscriber)

    def __get_middleware_callable(
        self, middleware: ConsumeMiddleware, consume_callable: ConsumeCallable
    ) -> ConsumeCallable:
        return lambda event, subscriber: middleware.on_consume(event, subscriber, consume_callable)



from buz.event.transactional_outbox import OutboxRecord
from buz.event.transactional_outbox.outbox_record_validation.outbox_record_validation_exception import (
    OutboxRecordValidationException,
)


class OutboxRecordSizeNotAllowedException(OutboxRecordValidationException):
    def __init__(self, record: OutboxRecord, size_limit_in_bytes: int, record_size: int):
        self.record = record
        self.size_limit_in_bytes = size_limit_in_bytes
        self.record_size = record_size
        super().__init__(
            error_message=(
                f"Record with id: {record.event_id} and fqn: {record.event_fqn} has size of "
                f"{record_size} bytes while the limit is {size_limit_in_bytes} bytes"
            )
        )



from buz import Handler


class HandlerAlreadyRegisteredException(Exception):
    def __init__(self, handler: Handler):
        self.handler = handler
        super().__init__("Handler has been registered and cannot be registered another time")



from __future__ import annotations

from logging import Logger
from typing import List, Optional, Union

from kafka import KafkaConsumer, KafkaProducer
from kafka.coordinator.assignors.range import RangePartitionAssignor

from buz.kafka.domain.services.kafka_admin_test_client import KafkaAdminTestClient
from buz.kafka.domain.services.kafka_consumer import DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING
from buz.kafka.infrastructure.kafka_python.kafka_poll_record import KafkaPollRecord
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.infrastructure.kafka_python.kafka_python_admin_client import KafkaPythonAdminClient

CONSUMER_POLL_TIMEOUT_MS = 1000


class KafkaPythonAdminTestClient(KafkaPythonAdminClient, KafkaAdminTestClient):
    def __init__(
        self,
        *,
        bootstrap_servers: List[str],
        client_id: str,
        logger: Logger,
        security_protocol: KafkaSupportedSecurityProtocols,
        sasl_mechanism: Union[str, None] = None,
        sasl_plain_username: Optional[str] = None,
        sasl_plain_password: Optional[str] = None,
    ):
        super().__init__(
            client_id=client_id,
            bootstrap_servers=bootstrap_servers,
            logger=logger,
            security_protocol=security_protocol,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_username=sasl_plain_username,
            sasl_plain_password=sasl_plain_password,
        )

    def send_message_to_topic(
        self,
        *,
        topic: str,
        message: bytes,
        headers: Optional[list[tuple[str, bytes]]] = None,
        partition_key: Optional[str] = None,
    ) -> None:
        producer = KafkaProducer(
            **self._generate_kafka_auth_configuration(),
        )

        producer.send(
            topic=topic,
            value=message,
            headers=headers,
            key=partition_key,
        )

        producer.close()

    def get_messages_from_topic(
        self,
        *,
        topic: str,
        consumer_group: str,
        max_number_of_messages_to_polling: int = DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING,
        offset: ConsumerInitialOffsetPosition = ConsumerInitialOffsetPosition.BEGINNING,
    ) -> list[KafkaPollRecord]:
        consumer = KafkaConsumer(
            **self._generate_kafka_auth_configuration(),
            group_id=consumer_group,
            enable_auto_commit=True,
            auto_offset_reset="earliest" if offset == ConsumerInitialOffsetPosition.BEGINNING else "latest",
            partition_assignment_strategy=(RangePartitionAssignor,),
        )

        consumer.subscribe(topics=topic)

        records = list(
            consumer.poll(
                timeout_ms=CONSUMER_POLL_TIMEOUT_MS,
                max_records=max_number_of_messages_to_polling,
            ).values()
        )

        consumer.close()

        topic_records = records[0] if len(records) > 0 else []

        return topic_records

    def delete_all_resources(
        self,
    ) -> None:
        self.delete_topics(topics=self.get_topics())
        self._wait_for_cluster_update()



from __future__ import annotations

from abc import ABC
from logging import Logger
from typing import Union

from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols


SSASL_MECHANISM = "SCRAM-SHA-512"
AUTHENTICATION_PROTOCOL_VALUES = Union[str, KafkaSupportedSecurityProtocols, None]


class KafkaPythonClientFactory(ABC):
    def __init__(
        self,
        logger: Logger,
        kafka_connection_config: KafkaConnectionConfig,
    ):
        self._logger = logger
        self._kafka_connection_config = kafka_connection_config

    def _get_autentication_configuration_by_security_protocol(self) -> dict[str, AUTHENTICATION_PROTOCOL_VALUES]:
        security_protocol_to_authentication_configuration: dict[str, dict[str, AUTHENTICATION_PROTOCOL_VALUES]] = {
            KafkaSupportedSecurityProtocols.SASL_PLAINTEXT.value: {
                "security_protocol": KafkaSupportedSecurityProtocols.SASL_PLAINTEXT,
                "sasl_mechanism": SSASL_MECHANISM,
                "sasl_plain_username": self._kafka_connection_config.user,
                "sasl_plain_password": self._kafka_connection_config.password,
            },
            KafkaSupportedSecurityProtocols.SASL_SSL.value: {
                "security_protocol": KafkaSupportedSecurityProtocols.SASL_SSL,
                "sasl_mechanism": SSASL_MECHANISM,
                "sasl_plain_username": self._kafka_connection_config.user,
                "sasl_plain_password": self._kafka_connection_config.password,
            },
        }

        configuration = security_protocol_to_authentication_configuration.get(
            self._kafka_connection_config.security_protocol.value
        )

        if configuration is not None:
            return configuration

        return {"security_protocol": KafkaSupportedSecurityProtocols.PLAINTEXT}



from typing import List

from buz.middleware.middleware_chain_builder import MiddlewareChainBuilder
from buz.query import Query, QueryResponse
from buz.query.asynchronous import QueryHandler
from buz.query.asynchronous.middleware.handle_middleware import HandleMiddleware, HandleCallable


class HandleMiddlewareChainResolver:
    def __init__(self, middlewares: List[HandleMiddleware]):
        self.__middlewares = middlewares
        self.__middleware_chain_builder: MiddlewareChainBuilder[
            HandleCallable, HandleMiddleware
        ] = MiddlewareChainBuilder(middlewares)

    async def resolve(self, query: Query, query_handler: QueryHandler, handle: HandleCallable) -> QueryResponse:
        chain_callable: HandleCallable = self.__middleware_chain_builder.get_chain_callable(
            handle, self.__get_middleware_callable
        )
        return await chain_callable(query, query_handler)

    @staticmethod
    def __get_middleware_callable(middleware: HandleMiddleware, next_callable: HandleCallable) -> HandleCallable:
        return lambda query, query_handler: middleware.on_handle(query, query_handler, next_callable)



from __future__ import annotations

import re
from logging import Logger
from typing import List, Optional, Union

from kafka import KafkaClient
from kafka.admin import KafkaAdminClient as NativeKafkaPythonAdminClient, NewTopic
from kafka.errors import TopicAlreadyExistsError

from buz.kafka.domain.exceptions.topic_already_created_exception import KafkaTopicsAlreadyCreatedException
from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.domain.models.kafka_topic import KafkaTopic
from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient


INTERNAL_KAFKA_TOPICS = {"__consumer_offsets", "_schema"}
PYTHON_KAFKA_DUPLICATED_TOPIC_ERROR_CODE = 36
CONSUMER_POLL_TIMEOUT_MS = 1000


class KafkaPythonAdminClient(KafkaAdminClient):
    def __init__(
        self,
        *,
        bootstrap_servers: List[str],
        client_id: str,
        logger: Logger,
        security_protocol: KafkaSupportedSecurityProtocols,
        sasl_mechanism: Union[str, None] = None,
        sasl_plain_username: Optional[str] = None,
        sasl_plain_password: Optional[str] = None,
    ):
        self._client_id = client_id
        self._bootstrap_servers = bootstrap_servers
        self._logger = bootstrap_servers
        self._sasl_mechanism = sasl_mechanism
        self._security_protocol = security_protocol
        self._sasl_plain_username = sasl_plain_username
        self._sasl_plain_password = sasl_plain_password

        self._kafka_admin = NativeKafkaPythonAdminClient(
            **self._generate_kafka_auth_configuration(),
        )

        self._kafka_client = KafkaClient(
            **self._generate_kafka_auth_configuration(),
        )

    def _generate_kafka_auth_configuration(self) -> dict:
        return {
            "client_id": self._client_id,
            "bootstrap_servers": self._bootstrap_servers,
            "security_protocol": self._security_protocol.value,
            "sasl_mechanism": self._sasl_mechanism,
            "sasl_plain_username": self._sasl_plain_username,
            "sasl_plain_password": self._sasl_plain_password,
        }

    def create_topics(
        self,
        *,
        topics: set[KafkaTopic],
    ) -> None:
        new_topics = [
            NewTopic(
                name=topic.name,
                num_partitions=topic.partitions,
                replication_factor=topic.replication_factor,
            )
            for topic in topics
        ]

        try:
            self._kafka_admin.create_topics(new_topics=new_topics)
        except TopicAlreadyExistsError as error:
            topic_names = self.__get_list_of_kafka_topics_from_topic_already_exists_error(error)
            raise KafkaTopicsAlreadyCreatedException(topic_names=topic_names)

    def __get_list_of_kafka_topics_from_topic_already_exists_error(self, error: TopicAlreadyExistsError) -> list[str]:
        message = str(error)
        response_message = re.search(r"topic_errors=\[.*?\]", message)
        topic_messages = re.findall(r"topic='[^']*', error_code=" + str(PYTHON_KAFKA_DUPLICATED_TOPIC_ERROR_CODE), response_message[0])  # type: ignore

        return [re.search("'.*'", topic_message)[0].strip("'") for topic_message in topic_messages]  # type: ignore

    def get_topics(
        self,
    ) -> set[str]:
        return set(self._kafka_admin.list_topics()) - INTERNAL_KAFKA_TOPICS

    def delete_topics(
        self,
        *,
        topics: set[str],
    ) -> None:
        self._kafka_admin.delete_topics(
            topics=topics,
        )

    def _wait_for_cluster_update(self) -> None:
        future = self._kafka_client.cluster.request_update()
        self._kafka_client.poll(future=future)



from abc import abstractmethod
from typing import Type

from buz import Handler
from buz.command.command import Command


class CommandHandler(Handler):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[Command]:
        pass

    @abstractmethod
    async def handle(self, command: Command) -> None:
        pass



from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Optional

from buz.kafka.domain.services.kafka_admin_client import KafkaAdminClient
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.infrastructure.kafka_python.kafka_poll_record import KafkaPollRecord

DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING = 999


class KafkaAdminTestClient(KafkaAdminClient, ABC):
    @abstractmethod
    def send_message_to_topic(
        self,
        *,
        topic: str,
        message: bytes,
        headers: Optional[list[tuple[str, bytes]]] = None,
        partition_key: Optional[str] = None,
    ) -> None:
        """
        This method was not though as a producer itself, it offers a way to send messages to a topic, that could be useful specially for test purposes
        """
        pass

    @abstractmethod
    def get_messages_from_topic(
        self,
        *,
        topic: str,
        consumer_group: str,
        max_number_of_messages_to_polling: int = DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING,
        offset: ConsumerInitialOffsetPosition = ConsumerInitialOffsetPosition.BEGINNING,
    ) -> list[KafkaPollRecord]:
        """
        This method is not though as a consumer itself, it offers a way to get messages from a topic, that could be useful specially for test purposes
        """
        pass

    @abstractmethod
    def delete_all_resources(
        self,
    ) -> None:
        """
        Only for test purposes, remove all the resources related to the kafka cluster, like topics, subscription groups
        """
        pass



from typing import cast

from buz.event import Event
from buz.event.transactional_outbox import FqnToEventMapper
from buz.event.transactional_outbox import OutboxRecord


class OutboxRecordToEventTranslator:
    def __init__(self, fqn_to_event_mapper: FqnToEventMapper):
        self.__fqn_to_event_mapper = fqn_to_event_mapper

    def translate(self, outbox_record: OutboxRecord) -> Event:
        event_klass = self.__fqn_to_event_mapper.get_message_klass_by_fqn(outbox_record.event_fqn)
        return cast(
            Event,
            event_klass.restore(
                id=outbox_record.event_id, created_at=outbox_record.parsed_created_at(), **outbox_record.event_payload
            ),
        )



from typing import Collection, Dict, List, Optional

from buz.kafka import (
    KafkaPythonProducer,
)
from buz.event import Event, EventBus
from buz.event.infrastructure.buz_kafka.publish_strategy.publish_strategy import KafkaPublishStrategy
from buz.event.middleware import (
    PublishMiddleware,
)
from buz.event.exceptions.event_not_published_exception import EventNotPublishedException
from buz.event.middleware.publish_middleware_chain_resolver import PublishMiddlewareChainResolver


class BuzKafkaEventBus(EventBus):
    def __init__(
        self,
        *,
        publish_strategy: KafkaPublishStrategy,
        publish_middlewares: Optional[List[PublishMiddleware]] = None,
        producer: KafkaPythonProducer,
    ):
        self.__publish_middleware_chain_resolver = PublishMiddlewareChainResolver(publish_middlewares or [])
        self.__publish_strategy = publish_strategy
        self.__producer = producer

    def publish(self, event: Event) -> None:
        self.__publish_middleware_chain_resolver.resolve(event, self.__perform_publish)

    def __perform_publish(self, event: Event) -> None:
        try:
            topic = self.__publish_strategy.get_topic(event)
            headers = self.__get_event_headers(event)
            self.__producer.produce(
                message=event,
                headers=headers,
                topic=topic,
            )
        except Exception as exc:
            raise EventNotPublishedException(event) from exc

    def bulk_publish(self, events: Collection[Event]) -> None:
        for event in events:
            self.publish(event)

    def __get_event_headers(self, event: Event) -> Dict:
        return {"event_fqn": event.fqn()}



from abc import abstractmethod, ABC
from typing import Optional

from kombu import Exchange


class PublishStrategy(ABC):
    @abstractmethod
    def get_exchange(self, event_fqn: str) -> Exchange:
        pass

    @abstractmethod
    def get_routing_key(self, event_fqn: str) -> Optional[str]:
        pass



from buz.query.asynchronous.base_query_handler import BaseQueryHandler
from buz.query.asynchronous.query_bus import QueryBus
from buz.query.asynchronous.query_handler import QueryHandler

__all__ = ["QueryBus", "BaseQueryHandler", "QueryHandler"]



from typing import Collection, Union

from buz.command.command import Command
from buz.command.synchronous import CommandHandler
from buz.command.asynchronous import CommandHandler as AsyncCommandHandler


class MoreThanOneCommandHandlerRelatedException(Exception):
    def __init__(self, command: Command, command_handlers: Collection[Union[CommandHandler, AsyncCommandHandler]]):
        self.command = command
        self.command_handlers = command_handlers
        super().__init__(f"There is more than one handler registered for {command.fqn()}.")



from buz.locator.locator import Locator
from buz.locator.handler_fqn_not_found_exception import HandlerFqnNotFoundException
from buz.locator.message_fqn_not_found_exception import MessageFqnNotFoundException

__all__ = ["Locator", "HandlerFqnNotFoundException", "MessageFqnNotFoundException"]



from abc import abstractmethod
from typing import Type

from buz import Handler
from buz.query import Query, QueryResponse


class QueryHandler(Handler):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[Query]:
        pass

    @abstractmethod
    async def handle(self, query: Query) -> QueryResponse:
        pass



from abc import abstractmethod
from typing import Type

from buz import Handler
from buz.event import Event


class Subscriber(Handler):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[Event]:
        pass

    @abstractmethod
    def consume(self, event: Event) -> None:
        pass



from typing import Dict


class EventRestoreException(Exception):
    def __init__(self, body: Dict, message: str) -> None:
        self.body = body
        self.message = message
        super().__init__(f"Event could not be restored. Body: {body}. Message: {message}")



from typing import Callable, Generic, TypeVar, List, Optional

from buz.middleware import Middleware

C = TypeVar("C", bound=Callable)
M = TypeVar("M", bound=Middleware)


class MiddlewareChainBuilder(Generic[C, M]):
    def __init__(self, middlewares: List[M]):
        self.__chain_callable: Optional[C] = None
        self.__middlewares = middlewares

    def get_chain_callable(self, base_case_callable: C, middleware_call: Callable[[M, C], C]) -> C:
        if self.__chain_callable is None:
            self.__chain_callable = self.__get_next_callable(0, base_case_callable, middleware_call)
        return self.__chain_callable

    def __get_next_callable(self, index: int, base_case_callable: C, middleware_call: Callable[[M, C], C]) -> C:
        if index == len(self.__middlewares):
            return base_case_callable
        next_callable = self.__get_next_callable(index + 1, base_case_callable, middleware_call)
        return middleware_call(self.__middlewares[index], next_callable)



from abc import ABC, abstractmethod
from typing import Type

from buz import Message


class Handler(ABC):
    @classmethod
    @abstractmethod
    def handles(cls) -> Type[Message]:
        pass

    @classmethod
    @abstractmethod
    def fqn(cls) -> str:
        pass



from buz.command.asynchronous.base_command_handler import BaseCommandHandler
from buz.command.asynchronous.command_bus import CommandBus
from buz.command.asynchronous.command_handler import CommandHandler

__all__ = ["CommandBus", "CommandHandler", "BaseCommandHandler"]



from abc import abstractmethod
from typing import Callable, Awaitable

from buz.middleware import Middleware
from buz.query import Query, QueryResponse
from buz.query.asynchronous import QueryHandler

HandleCallable = Callable[[Query, QueryHandler], Awaitable[QueryResponse]]


class HandleMiddleware(Middleware):
    @abstractmethod
    async def on_handle(self, query: Query, query_handler: QueryHandler, handle: HandleCallable) -> QueryResponse:
        pass



from logging import Logger
from typing import Union

from buz.event import EventBus
from buz.event.transactional_outbox import (
    OutboxRepository,
    OutboxRecordToEventTranslator,
    OutboxRecordStreamFinder,
    OutboxRecord,
)
from buz.locator import MessageFqnNotFoundException
from buz.event.exceptions.event_not_published_exception import EventNotPublishedException


class TransactionalOutboxWorker:
    def __init__(
        self,
        outbox_repository: OutboxRepository,
        outbox_record_stream_finder: OutboxRecordStreamFinder,
        outbox_record_to_event_translator: OutboxRecordToEventTranslator,
        event_bus: EventBus,
        logger: Logger,
        max_retries: int,
    ):
        self.__outbox_repository = outbox_repository
        self.__outbox_record_stream_finder = outbox_record_stream_finder
        self.__outbox_record_to_event_translator = outbox_record_to_event_translator
        self.__event_bus = event_bus
        self.__logger = logger
        self.__max_retries = max_retries

    def start(self) -> None:
        for outbox_record in self.__outbox_record_stream_finder.find():
            try:
                event = self.__outbox_record_to_event_translator.translate(outbox_record)

                self.__event_bus.publish(event)
                outbox_record.mark_as_delivered()

            except (MessageFqnNotFoundException, EventNotPublishedException) as e:
                self.__mark_delivery_error(outbox_record)
                self.__log_first_time_exception(e, outbox_record)

            except Exception as e:
                self.__logger.exception(e)
                self.__mark_delivery_error(outbox_record)

            self.__outbox_repository.save(outbox_record)

    def __mark_delivery_error(self, outbox_record: OutboxRecord) -> None:
        outbox_record.mark_delivery_error()
        if outbox_record.delivery_errors >= self.__max_retries:
            outbox_record.pause_delivery()

    def __log_first_time_exception(
        self, exc: Union[MessageFqnNotFoundException, EventNotPublishedException], outbox_record: OutboxRecord
    ) -> None:
        if outbox_record.delivery_errors == 1:
            self.__logger.exception(exc)



from abc import abstractmethod
from typing import Callable

from buz.middleware import Middleware
from buz.query import Query, QueryResponse
from buz.query.synchronous import QueryHandler

HandleCallable = Callable[[Query, QueryHandler], QueryResponse]


class HandleMiddleware(Middleware):
    @abstractmethod
    def on_handle(self, query: Query, query_handler: QueryHandler, handle: HandleCallable) -> QueryResponse:
        pass



from kombu import Queue

from buz.event.infrastructure.kombu.consume_strategy.consume_strategy import ConsumeStrategy
from buz.event.infrastructure.kombu.publish_strategy import PublishStrategy
from buz.locator import Locator


class QueuePerSubscriberConsumeStrategy(ConsumeStrategy):
    def __init__(self, publish_strategy: PublishStrategy, locator: Locator):
        self.__publish_strategy = publish_strategy
        self.__locator = locator

    def get_queue(self, subscriber_fqn: str) -> Queue:
        event_fqn = self.__locator.get_handler_by_fqn(subscriber_fqn).handles().fqn()
        exchange = self.__publish_strategy.get_exchange(event_fqn)
        key = self.__publish_strategy.get_routing_key(event_fqn)
        return Queue(subscriber_fqn, exchange=exchange, key=key)



from datetime import datetime
from typing import Dict, Optional, Any, ClassVar
from uuid import UUID
from dataclasses import dataclass, fields


@dataclass
class OutboxRecord:  # type: ignore[misc]
    DATE_TIME_FORMAT: ClassVar[str] = "%Y-%m-%d %H:%M:%S.%f"

    event_id: UUID
    event_fqn: str
    event_payload: Dict[str, Any]  # type: ignore[misc]
    created_at: datetime
    delivered_at: Optional[datetime] = None
    delivery_errors: int = 0
    delivery_paused_at: Optional[datetime] = None

    def parsed_created_at(self) -> str:
        return self.created_at.strftime(self.DATE_TIME_FORMAT)

    def mark_delivery_error(self) -> None:
        self.delivered_at = None
        self.delivery_errors += 1

    def mark_as_delivered(self) -> None:
        self.delivered_at = datetime.now()

    def pause_delivery(self) -> None:
        self.delivery_paused_at = datetime.now()

    def get_attrs(self) -> Dict[str, Any]:
        attrs = {}
        for field in fields(self):
            property_name = field.name
            property_value = getattr(self, property_name)
            attrs[property_name] = property_value

        return attrs



from logging import Logger

from buz.kafka.domain.models.kafka_connection_config import KafkaConnectionConfig
from buz.kafka.infrastructure.kafka_python.factories.kafka_python_client_factory import KafkaPythonClientFactory
from buz.kafka.infrastructure.kafka_python.kafka_python_producer import KafkaPythonProducer
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import JSONByteSerializer


class KafkaPythonProducerFactory(KafkaPythonClientFactory):
    def __init__(
        self,
        logger: Logger,
        kafka_connection_config: KafkaConnectionConfig,
        byte_serializer: ByteSerializer = JSONByteSerializer(),
    ):
        super().__init__(logger=logger, kafka_connection_config=kafka_connection_config)
        self._kafka_connection_config = kafka_connection_config
        self._byte_serializer = byte_serializer

    def build(self) -> KafkaPythonProducer:
        return KafkaPythonProducer(
            bootstrap_servers=self._kafka_connection_config.bootstrap_servers,
            client_id=self._kafka_connection_config.client_id,
            logger=self._logger,
            byte_serializer=self._byte_serializer,
            **self._get_autentication_configuration_by_security_protocol(),  # type: ignore
        )



from abc import abstractmethod, ABC
from typing import List
from buz.event.subscriber import Subscriber


class KafkaConsumeStrategy(ABC):
    @abstractmethod
    def get_topics(self, subscriber: Subscriber) -> List[str]:
        pass

    @abstractmethod
    def get_subscription_group(self, subscriber: Subscriber) -> str:
        pass



from typing import List, Optional

from buz.locator import Locator
from buz.query import Query, QueryResponse
from buz.query.asynchronous import QueryBus
from buz.query.asynchronous import QueryHandler
from buz.query.asynchronous.middleware import HandleMiddleware, HandleMiddlewareChainResolver
from buz.query import MoreThanOneQueryHandlerRelatedException


class SelfProcessQueryBus(QueryBus):
    def __init__(
        self,
        locator: Locator[Query, QueryHandler],
        middlewares: Optional[List[HandleMiddleware]] = None,
    ):
        self.__locator = locator
        self.__handle_middleware_chain_resolver = HandleMiddlewareChainResolver(middlewares or [])

    async def handle(self, query: Query) -> QueryResponse:
        query_handlers = self.__locator.get(query)

        if len(query_handlers) > 1:
            raise MoreThanOneQueryHandlerRelatedException(query, query_handlers)

        query_handler = query_handlers[0]
        return await self.__handle_middleware_chain_resolver.resolve(query, query_handler, self.__perform_handle)

    @staticmethod
    async def __perform_handle(query: Query, query_handler: QueryHandler) -> QueryResponse:
        return await query_handler.handle(query)



class HandlerFqnNotFoundException(Exception):
    def __init__(self, handler_fqn: str):
        self.handler_fqn = handler_fqn
        super().__init__(f"Handler with fqn {handler_fqn} has not been found")



from typing import Collection, Optional

from buz.event import Event, EventBus
from buz.event.transactional_outbox.event_to_outbox_record_translator import EventToOutboxRecordTranslator
from buz.event.transactional_outbox.outbox_record_validation.outbox_record_validator import OutboxRecordValidator
from buz.event.transactional_outbox.outbox_repository import OutboxRepository


class TransactionalOutboxEventBus(EventBus):
    def __init__(
        self,
        outbox_repository: OutboxRepository,
        event_to_outbox_record_translator: EventToOutboxRecordTranslator,
        outbox_record_validator: Optional[OutboxRecordValidator] = None,
    ):
        self.__outbox_repository = outbox_repository
        self.__event_to_outbox_record_translator = event_to_outbox_record_translator
        self.__outbox_record_validator = outbox_record_validator

    def publish(self, event: Event) -> None:
        """
        Raises:
            OutboxRecordValidationException: If any validation inside outbox_record_validator fails.
        """
        outbox_record = self.__event_to_outbox_record_translator.translate(event)
        if self.__outbox_record_validator is not None:
            self.__outbox_record_validator.validate(record=outbox_record)

        self.__outbox_repository.save(outbox_record)

    def bulk_publish(self, events: Collection[Event]) -> None:
        for event in events:
            self.publish(event)



from time import sleep
from typing import Generator

from buz.event.transactional_outbox import OutboxCriteriaFactory
from buz.event.transactional_outbox.outbox_record import OutboxRecord
from buz.event.transactional_outbox.outbox_repository import OutboxRepository
from buz.event.transactional_outbox.outbox_record_finder import OutboxRecordStreamFinder


class PollingOutboxRecordStreamFinder(OutboxRecordStreamFinder):
    def __init__(
        self,
        outbox_repository: OutboxRepository,
        criteria_factory: OutboxCriteriaFactory,
        find_iteration_sleep_seconds: int = 2,
    ):
        self.__outbox_repository = outbox_repository
        self.__criteria_factory = criteria_factory
        self.__find_iteration_sleep_seconds = find_iteration_sleep_seconds

    def find(self) -> Generator[OutboxRecord, None, None]:
        while True:
            criteria = self.__criteria_factory.create()
            for record in self.__outbox_repository.find(criteria=criteria):
                yield record

            sleep(self.__find_iteration_sleep_seconds)



class TermSignalInterruptionException(Exception):
    def __init__(self):
        self.message = "Stopping consumer due to TERM Signal, subscriber had not yet started processing the event"
        super(TermSignalInterruptionException, self).__init__(self.message)



from buz.command.synchronous import CommandBus
from buz.command import Command
from buz.command.asynchronous import CommandBus as AsynchronousCommandbus

from buz.wrapper import AsyncToSync


class SyncedAsyncCommandBus(CommandBus):
    def __init__(self, async_bus: AsynchronousCommandbus):
        self.__wrapped_handler = AsyncToSync[Command, None](async_bus.handle)

    def handle(self, command: Command) -> None:
        self.__wrapped_handler(command)



from typing import Collection, List, Optional

from buz.event import Event, EventBus, Subscriber
from buz.event.middleware import (
    PublishMiddleware,
    ConsumeMiddleware,
    PublishMiddlewareChainResolver,
    ConsumeMiddlewareChainResolver,
)
from buz.locator import Locator


class SyncEventBus(EventBus):
    def __init__(
        self,
        locator: Locator[Event, Subscriber],
        publish_middlewares: Optional[List[PublishMiddleware]] = None,
        consume_middlewares: Optional[List[ConsumeMiddleware]] = None,
    ):
        self.__locator = locator
        self.__publish_middleware_chain_resolver = PublishMiddlewareChainResolver(publish_middlewares or [])
        self.__consume_middleware_chain_resolver = ConsumeMiddlewareChainResolver(consume_middlewares or [])

    def publish(self, event: Event) -> None:
        self.__publish_middleware_chain_resolver.resolve(event, self.__perform_publish)

    def __perform_publish(self, event: Event) -> None:
        subscribers = self.__locator.get(event)
        for subscriber in subscribers:
            self.__consume_middleware_chain_resolver.resolve(event, subscriber, self.__perform_consume)

    def __perform_consume(self, event: Event, subscriber: Subscriber) -> None:
        subscriber.consume(event)

    def bulk_publish(self, events: Collection[Event]) -> None:
        for event in events:
            self.publish(event)



from buz import Handler


class HandlerNotRegisteredException(Exception):
    def __init__(self, handler: Handler):
        self.handler = handler
        super().__init__("Handler has not been registered and cannot be unregistered")



from abc import ABC
from dataclasses import field, dataclass
from datetime import datetime
from inspect import signature, Parameter
from types import MappingProxyType
from typing import Any, ClassVar, Dict, get_origin, get_args, Union

from uuid_utils.compat import uuid7


@dataclass(frozen=True)
class Message(ABC):
    DATE_TIME_FORMAT: ClassVar[str] = "%Y-%m-%d %H:%M:%S.%f"

    id: str = field(init=False, default_factory=lambda: str(uuid7()))
    created_at: str = field(
        init=False, default_factory=lambda: datetime.strftime(datetime.now(), Message.DATE_TIME_FORMAT)
    )

    @classmethod
    def fqn(cls) -> str:
        return f"{cls.__module__}.{cls.__name__}"

    @classmethod
    def restore(cls, **kwargs: Any) -> "Message":  # type: ignore[misc]
        message_id = kwargs.pop("id")
        created_at = kwargs.pop("created_at")

        instance = cls.__from_dict(kwargs)  # type: ignore

        object.__setattr__(instance, "id", message_id)
        object.__setattr__(instance, "created_at", created_at)
        return instance

    @classmethod
    def __from_dict(cls, data: Dict[str, Any]) -> "Message":
        expected_params = signature(cls).parameters
        actual_params = {key: value for key, value in data.items() if key in expected_params}
        if len(actual_params) == len(expected_params):
            return cls(**actual_params)  # type: ignore[call-arg]

        filled_optional_attributes = cls.__fill_optional_attributes(actual_params, expected_params)
        return cls(**filled_optional_attributes, **actual_params)  # type: ignore[call-arg]

    @classmethod
    def __fill_optional_attributes(
        cls, actual_params: Dict[str, Any], expected_params: MappingProxyType
    ) -> Dict[str, None]:
        filled_optional_attributes: Dict[str, None] = {}
        for key, parameter in expected_params.items():
            if key not in actual_params and cls.__is_optional(parameter) is True:
                filled_optional_attributes[key] = None
        return filled_optional_attributes

    @classmethod
    def __is_optional(cls, parameter: Parameter) -> bool:
        return get_origin(parameter.annotation) is Union and any(
            (arg is type(None) for arg in get_args(parameter.annotation))  # noqa E721
        )

    def parsed_created_at(self) -> datetime:
        return datetime.strptime(self.created_at, self.DATE_TIME_FORMAT)



from abc import ABC, abstractmethod
from typing import List

from buz.event import Event, Subscriber


class RejectCallback(ABC):
    @abstractmethod
    def on_reject(self, event: Event, subscribers: List[Subscriber]) -> None:
        pass



from logging import Logger
from typing import Optional


class ContainerLocatorResolutionConfiguration:
    def __init__(self, allow_partial_resolve: bool, logger: Optional[Logger] = None):
        if allow_partial_resolve is True and logger is None:
            raise ValueError("Logger cannot be None with partial resolution enabled")

        self.allow_partial_resolve = allow_partial_resolve
        self.logger = logger



from buz.command.synchronous.command_handler import CommandHandler
from buz.command.synchronous.command_bus import CommandBus
from buz.command.synchronous.base_command_handler import BaseCommandHandler

__all__ = ["CommandHandler", "BaseCommandHandler", "CommandBus"]



from collections import defaultdict
from threading import Lock
from typing import DefaultDict, Generic, List, Optional, Set, Type, TypeVar, cast

from pypendency.container import AbstractContainer
from pypendency.exceptions import ServiceNotFoundInContainer

from buz import Handler
from buz import Message
from buz.locator import HandlerFqnNotFoundException, Locator, MessageFqnNotFoundException
from buz.locator.pypendency import HandlerNotFoundException
from buz.locator.pypendency import HandlerNotRegisteredException
from buz.locator.pypendency.container_locator_resolution_configuration import ContainerLocatorResolutionConfiguration

K = TypeVar("K", bound=Message)
V = TypeVar("V", bound=Handler)
MessageFqn = str


class ContainerLocator(Locator, Generic[K, V]):
    CHECK_MODE_REGISTER_TIME = "register"
    CHECK_MODE_GET_TIME = "get"

    def __init__(
        self,
        container: AbstractContainer,
        check_mode: str = CHECK_MODE_REGISTER_TIME,
        container_locator_resolution_configuration: Optional[ContainerLocatorResolutionConfiguration] = None,
    ) -> None:
        self.__container = container
        self.__check_mode = check_mode
        self.__allow_partial_resolve = None
        self.__logger = None
        if container_locator_resolution_configuration is not None:
            self.__allow_partial_resolve = container_locator_resolution_configuration.allow_partial_resolve
            self.__logger = container_locator_resolution_configuration.logger
        self.__mapping: DefaultDict[MessageFqn, List[V]] = defaultdict(list)
        self.__handler_ids: Set[str] = set()
        self.__handlers_resolved = False
        self.__lock = Lock()

    def register(self, handler_id: str) -> None:
        if handler_id in self.__handler_ids:
            return
        if self.__check_mode == self.CHECK_MODE_REGISTER_TIME:
            self.__guard_handler_not_found(handler_id)
        self.__handler_ids.add(handler_id)
        self.__handlers_resolved = False

    def __guard_handler_not_found(self, handler_id: str) -> None:
        if not self.__container.has(handler_id):
            raise HandlerNotFoundException(handler_id)

    def unregister(self, handler_id: str) -> None:
        self.__guard_handler_not_registered(handler_id)
        self.__handler_ids.remove(handler_id)
        self.__handlers_resolved = False

    def __guard_handler_not_registered(self, handler_id: str) -> None:
        if handler_id not in self.__handler_ids:
            raise HandlerNotRegisteredException(handler_id)

    def get(self, message: K) -> List[V]:
        self.__ensure_handlers_resolved()
        return self.__mapping.get(message.fqn(), [])

    def __ensure_handlers_resolved(self) -> None:
        self.__lock.acquire()
        try:
            if self.__handlers_resolved is False:
                self._resolve_handlers()
        finally:
            self.__lock.release()

    def _resolve_handlers(self) -> None:
        self.__mapping = defaultdict(list)
        for handler_id in self.__handler_ids:
            if self.__check_mode == self.CHECK_MODE_GET_TIME:
                self.__guard_handler_not_found(handler_id)
            try:
                handler: V = self.__container.get(handler_id)
                message_fqn = handler.handles().fqn()
                self.__mapping[message_fqn].append(handler)
            except ServiceNotFoundInContainer as e:
                if self.__allow_partial_resolve is False:
                    raise e
                if self.__logger is not None:
                    self.__logger.error(f"Error while resolving handler {handler_id}: {e}")
        self.__handlers_resolved = True

    def get_handler_by_fqn(self, handler_fqn: str) -> V:
        self.__ensure_handlers_resolved()
        for message_handlers in self.__mapping.values():
            for handler in message_handlers:
                if handler_fqn == handler.fqn():
                    return handler
        raise HandlerFqnNotFoundException(handler_fqn)

    def get_message_klass_by_fqn(self, message_fqn: str) -> Type[K]:
        self.__ensure_handlers_resolved()
        try:
            handler = self.__mapping.get(message_fqn, [])[0]
            return cast(Type[K], handler.handles())
        except (IndexError, TypeError, HandlerFqnNotFoundException):
            raise MessageFqnNotFoundException(message_fqn)



from __future__ import annotations

from logging import Logger
from typing import Callable, List, Type, Tuple, Union, cast

from kafka import KafkaConsumer, TopicPartition, OffsetAndMetadata
from kafka.coordinator.assignors.abstract import AbstractPartitionAssignor
from kafka.coordinator.assignors.range import RangePartitionAssignor

from buz.kafka.domain.models.kafka_supported_security_protocols import KafkaSupportedSecurityProtocols
from buz.kafka.infrastructure.kafka_python.exception.consumer_interrupted_exception import (
    ConsumerInterruptedException,
)
from buz.kafka.infrastructure.serializers.byte_serializer import ByteSerializer
from buz.kafka.infrastructure.serializers.implementations.json_byte_serializer import JSONByteSerializer
from buz.kafka.infrastructure.serializers.kafka_header_serializer import KafkaHeaderSerializer
from buz.kafka.domain.models.kafka_consumer_record import KafkaConsumerRecord
from buz.kafka.infrastructure.kafka_python.kafka_poll_record import KafkaPollRecord
from buz.kafka.domain.models.consumer_initial_offset_position import ConsumerInitialOffsetPosition
from buz.kafka.domain.services.kafka_consumer import (
    DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING,
)


CONSUMER_POLL_TIMEOUT_MS = 1000
# https://kafka-python.readthedocs.io/en/master/apidoc/KafkaConsumer.html#kafka.KafkaConsumer.poll
SESSION_TIMEOUT_MS = 10000
# https://docs.confluent.io/platform/current/installation/configuration/consumer-configs.html#session-timeout-ms


class KafkaPythonConsumer(KafkaConsumer):
    def __init__(
        self,
        *,
        client_id: str,
        bootstrap_servers: List[str],
        logger: Logger,
        security_protocol: KafkaSupportedSecurityProtocols,
        sasl_mechanism: Union[str, None] = None,
        sasl_plain_username: Union[str, None] = None,
        sasl_plain_password: Union[str, None] = None,
        partition_assignors: Tuple[Type[AbstractPartitionAssignor], ...] = (RangePartitionAssignor,),
        byte_serializer: ByteSerializer = JSONByteSerializer(),
    ):
        self._client_id = client_id
        self._bootstrap_servers = bootstrap_servers
        self._logger = logger
        self._sasl_mechanism = sasl_mechanism
        self._security_protocol = security_protocol
        self._sasl_plain_username = sasl_plain_username
        self._sasl_plain_password = sasl_plain_password
        self._partition_assignors = partition_assignors
        self.__byte_serializer = byte_serializer
        self.__header_serializer = KafkaHeaderSerializer()

        self._gracefully_stop = False

    def consume(
        self,
        *,
        topics: List[str],
        consumer_group: str,
        consumption_callback: Callable[[KafkaConsumerRecord], None],
        initial_offset_position: ConsumerInitialOffsetPosition = ConsumerInitialOffsetPosition.BEGINNING,
        number_of_messages_to_polling: int = DEFAULT_NUMBER_OF_MESSAGES_TO_POLLING,
    ) -> None:
        self._gracefully_stop = False

        consumer = self.__generate_consumer(
            consumer_group=consumer_group,
            initial_offset_position=initial_offset_position,
        )

        topics_pattern = "|".join(topics)

        consumer.subscribe(topics=topics_pattern)

        while True:
            try:
                if self._gracefully_stop is True:
                    raise ConsumerInterruptedException("The consumer execution was interrupted")

                consumer_metadata: dict[str, str] = {}
                poll_results = consumer.poll(
                    timeout_ms=CONSUMER_POLL_TIMEOUT_MS,
                    max_records=number_of_messages_to_polling,
                )

                for topic_partition, consumer_records in poll_results.items():
                    for consumer_record in consumer_records:
                        kafka_poll_record = cast(KafkaPollRecord, consumer_record)
                        if kafka_poll_record.value is not None:
                            consumption_callback(
                                KafkaConsumerRecord(
                                    value=self.__byte_serializer.unserialize(kafka_poll_record.value),
                                    headers=self.__header_serializer.unserialize(kafka_poll_record.headers),
                                )
                            )

                        self.__mark_record_as_consumed(consumer_metadata, kafka_poll_record)

                        if self._gracefully_stop is True:
                            raise ConsumerInterruptedException("The consumer execution was interrupted")

                consumer.commit()

            except ConsumerInterruptedException:
                self.__close_consumer(consumer, consumer_metadata)
                return
            except Exception:
                self.__close_consumer(consumer, consumer_metadata)
                raise

    def __generate_consumer(
        self,
        *,
        consumer_group: str,
        initial_offset_position: ConsumerInitialOffsetPosition,
    ) -> KafkaConsumer:
        auto_offset_rest = (
            "earliest" if initial_offset_position == ConsumerInitialOffsetPosition.BEGINNING else "latest"
        )

        return KafkaConsumer(
            bootstrap_servers=self._bootstrap_servers,
            security_protocol=self._security_protocol.value,
            sasl_mechanism=self._sasl_mechanism,
            sasl_plain_username=self._sasl_plain_username,
            sasl_plain_password=self._sasl_plain_password,
            client_id=self._client_id,
            group_id=consumer_group,
            enable_auto_commit=False,
            auto_offset_reset=auto_offset_rest,
            session_timeout_ms=SESSION_TIMEOUT_MS,
            partition_assignment_strategy=list(self._partition_assignors),
        )

    def __mark_record_as_consumed(self, consumer_metadata: dict[str, str], consumer_record: KafkaPollRecord) -> None:
        """
        the committed offset should be
        the next message your application should consume, i.e.: last_offset + 1.
        """
        next_offset_to_be_consumed = consumer_record.offset + 1
        consumer_metadata[
            TopicPartition(topic=consumer_record.topic, partition=consumer_record.partition)
        ] = OffsetAndMetadata(next_offset_to_be_consumed, "")

    def __close_consumer(self, consumer: KafkaConsumer, consumer_metadata: dict) -> None:
        consumer.commit(offsets=consumer_metadata)

        self._logger.info("Closing Consumer's connection")
        consumer.close(autocommit=False)

    def request_stop(self) -> None:
        self._gracefully_stop = True



from abc import ABC, abstractmethod
from typing import List

from buz.event import Event, Subscriber


class ConsumeRetrier(ABC):
    @abstractmethod
    def should_retry(self, event: Event, subscribers: List[Subscriber]) -> bool:
        pass

    @abstractmethod
    def register_retry(self, event: Event, subscribers: List[Subscriber]) -> None:
        pass
